class CursorWrapper:
    def __init__(self, cursor):
        self.cursor = cursor

    def _fix_for_params(self, query, params):
        if query.endswith(';'):
            query = query[:-1]
        if params is None:
            params = []
        elif hasattr(params, 'keys'):
            # Handle params as dict
            args = {k: "?" % k for k in params}
            query = query % args
        else:
            # Handle params as sequence
            args = ['?' for i in range(len(params))]
            query = query % tuple(args)
        return query, list(params)

    def execute(self, query, params=None):
        self.times = 0
        query, params = self._fix_for_params(query, params)
        # print(query, params)
        return self.cursor.execute(query, params)

    def executemany(self, query, params=None):
        self.times = 0
        query, params = self._fix_for_params(query, params)
        return self.cursor.executemany(query, params)

    def close(self):
        try:
            self.cursor.close()
        except:
            # already closed
            pass

    def __getattr__(self, attr):
        return getattr(self.cursor, attr)

    def __iter__(self):
        return iter(self.cursor)

    def fetchall(self):
        rows = self.cursor.fetchall()
        rows = [tuple(r) for r in rows]
        return rows

    def fetchmany(self, size=None):
        # workaround for endless loop
        if self.times > 0:
            return []
        self.times += 1
        rows = self.cursor.fetchmany(size)
        rows = [tuple(r) for r in rows]
        return rows

    def fetchone(self):
        row = self.cursor.fetchone()
        return tuple(row) if row else None



from collections import namedtuple

from django.db.backends.base.introspection import (
    BaseDatabaseIntrospection, FieldInfo as BaseFieldInfo, TableInfo,
)
from django.db.models import Index
from django.utils.datastructures import OrderedSet

FieldInfo = namedtuple(
    'FieldInfo', BaseFieldInfo._fields + ('auto_increment', ))


class DatabaseIntrospection(BaseDatabaseIntrospection):
    data_types_reverse = {
        'bigint': 'BigIntegerField',
        'varchar': 'CharField',
        'integer': 'IntegerField',
        'bit': 'BooleanField',
        'date': 'DateField',
        'timestamp': 'DateTimeField',
        'numeric': 'IntegerField',
        'double': 'FloatField',
        'varbinary': 'BinaryField',
        'longvarchar': 'TextField',
        'longvarbinary': 'BinaryField',
        'time': 'TimeField',
        'smallint': 'SmallIntegerField',
        'tinyint': 'SmallIntegerField',
    }

    def get_field_type(self, data_type, field_info):
        field_type = super().get_field_type(data_type, field_info)
        if (field_info.auto_increment and
                field_type in {'BigIntegerField', 'IntegerField', 'SmallIntegerField'}):
            return 'AutoField'
        return field_type

    def get_sequences(self, cursor, table_name, table_fields=()):
        """
        Return a list of introspected sequences for table_name. Each sequence
        is a dict: {'table': <table_name>, 'column': <column_name>}. An optional
        'name' key can be added if the backend supports named sequences.
        """
        for field_info in self.get_table_description(cursor, table_name):
            if 'auto_increment' in field_info and field_info['auto_increment']:
                return [{'table': table_name, 'column': field_info.name}]
        return []

    def get_table_list(self, cursor):
        cursor.execute("""
          SELECT TABLE_NAME
          FROM information_schema.tables
          WHERE TABLE_SCHEMA = 'SQLUser'
        """)
        return [TableInfo(row[0], 't') for row in cursor.fetchall()]

    def get_table_description(self, cursor, table_name):
        """
        Return a description of the table with the DB-API cursor.description
        interface.
        """

        cursor.execute("""
            SELECT 
                COLUMN_NAME,
                DATA_TYPE,
                CHARACTER_MAXIMUM_LENGTH,
                NUMERIC_PRECISION,
                NUMERIC_SCALE,
                IS_NULLABLE,
                AUTO_INCREMENT
            FROM INFORMATION_SCHEMA.COLUMNS 
            WHERE TABLE_SCHEMA = %s
            AND TABLE_NAME = %s
            ORDER BY ORDINAL_POSITION
        """,
                       ['SQLUser', table_name]
                       )

        description = [
            # name type_code display_size internal_size precision scale null_ok default collation
            # auto_increment
            FieldInfo(
                name,
                data_type,
                None,
                length,
                precision,
                scale,
                isnull == 'YES',
                '',
                '',
                auto_increment == 'YES',
            )
            for name, data_type, length, precision, scale, isnull, auto_increment in cursor.fetchall()
        ]
        return description

    def get_relations(self, cursor, table_name):
        """
        Return a dictionary of {field_name: (field_name_other_table, other_table)}
        representing all relationships to the given table.
        """
        # Dictionary of relations to return
        cursor.execute("""
            SELECT column_name, referenced_column_name, referenced_table_name
            FROM information_schema.key_column_usage
            WHERE table_schema = %s
                AND table_name = %s
                AND referenced_table_name IS NOT NULL
                AND referenced_column_name IS NOT NULL
        """, ['SQLUser', table_name])
        return {
            field_name: (other_field, other_table)
            for field_name, other_field, other_table in cursor.fetchall()
        }

    # def get_primary_key_column(self, cursor, table_name):
    #     """
    #     Return the name of the primary key column for the given table.
    #     """
        # return super().get_primary_key_column(cursor, table_name)

    def get_constraints(self, cursor, table_name):
        """
        Retrieve any constraints or keys (unique, pk, fk, check, index)
        across one or more columns.

        Return a dict mapping constraint names to their attributes,
        where attributes is a dict with keys:
         * columns: List of columns this covers
         * primary_key: True if primary key, False otherwise
         * unique: True if this is a unique constraint, False otherwise
         * foreign_key: (table, column) of target, or None
         * check: True if check constraint, False otherwise
         * index: True if index, False otherwise.
         * orders: The order (ASC/DESC) defined for the columns of indexes
         * type: The type of the index (btree, hash, etc.)

        Some backends may return special constraint names that don't exist
        if they don't name constraints of a certain type (e.g. SQLite)
        """
        constraints = {}
        # Get the actual constraint names and columns
        name_query = """
            SELECT kc.constraint_name, kc.column_name,
                kc.referenced_table_name, kc.referenced_column_name,
                c.constraint_type
            FROM
                information_schema.key_column_usage AS kc,
                information_schema.table_constraints AS c
            WHERE
                kc.table_schema = %s AND
                c.table_schema = kc.table_schema AND
                c.constraint_name = kc.constraint_name AND
                c.constraint_type != 'CHECK' AND
                kc.table_name = %s
            ORDER BY kc.ordinal_position
        """
        cursor.execute(name_query, ['SQLUser', table_name])
        for constraint, column, ref_table, ref_column, kind in cursor.fetchall():
            if constraint not in constraints:
                constraints[constraint] = {
                    'columns': OrderedSet(),
                    'primary_key': kind == 'PRIMARY KEY',
                    'unique': kind in {'PRIMARY KEY', 'UNIQUE'},
                    'index': False,
                    'check': False,
                    'foreign_key': (ref_table, ref_column) if ref_column else None,
                }
                if self.connection.features.supports_index_column_ordering:
                    constraints[constraint]['orders'] = []
            constraints[constraint]['columns'].add(column)
        # Add check constraints.
        if self.connection.features.can_introspect_check_constraints:
            unnamed_constraints_index = 0
            columns = {info.name for info in self.get_table_description(
                cursor, table_name)}
            type_query = """
                SELECT cc.constraint_name, cc.check_clause
                FROM
                    information_schema.check_constraints AS cc,
                    information_schema.table_constraints AS tc
                WHERE
                    cc.constraint_schema = %s AND
                    tc.table_schema = cc.constraint_schema AND
                    cc.constraint_name = tc.constraint_name AND
                    tc.constraint_type = 'CHECK' AND
                    tc.table_name = %s
            """
            cursor.execute(type_query, ['SQLUser', table_name])
            for constraint, check_clause in cursor.fetchall():
                constraint_columns = self._parse_constraint_columns(
                    check_clause, columns)
                # Ensure uniqueness of unnamed constraints. Unnamed unique
                # and check columns constraints have the same name as
                # a column.
                if set(constraint_columns) == {constraint}:
                    unnamed_constraints_index += 1
                    constraint = '__unnamed_constraint_%s__' % unnamed_constraints_index
                constraints[constraint] = {
                    'columns': constraint_columns,
                    'primary_key': False,
                    'unique': False,
                    'index': False,
                    'check': True,
                    'foreign_key': None,
                }

        index_query = """
            SELECT 
                INDEX_NAME, 
                COLUMN_NAME, 
                PRIMARY_KEY,
                NON_UNIQUE,
                ASC_OR_DESC
            FROM INFORMATION_SCHEMA. INDEXES
            WHERE TABLE_SCHEMA = %s
              AND TABLE_NAME = %s
            ORDER BY ORDINAL_POSITION
        """
        cursor.execute(index_query, ['SQLUser', table_name])
        for index, column, primary, non_unique, order in cursor.fetchall():
            if index not in constraints:
                constraints[index] = {
                    'columns': OrderedSet(),
                    'primary_key': primary == 1,
                    'unique': not non_unique,
                    'check': False,
                    'foreign_key': None,
                    'orders': [],
                }
                constraints[index]['columns'].add(column)
                constraints[index]['orders'].append(
                    'DESC' if order == 'D' else 'ASC')

        # Convert the sorted sets to lists
        for constraint in constraints.values():
            constraint['columns'] = list(constraint['columns'])
        return constraints



from django.db.backends.base.validation import BaseDatabaseValidation

class DatabaseValidation(BaseDatabaseValidation):
    def check(self, **kwargs):
        return []

    def check_field(self, field, **kwargs):
        return []



from django.db.backends.base.features import BaseDatabaseFeatures
class DatabaseFeatures(BaseDatabaseFeatures):
    empty_fetchmany_value = []

    requires_literal_defaults = True
    
    supports_transactions = False
    uses_savepoints = False
    has_bulk_insert =  False
    has_native_uuid_field =  False
    supports_timezones = False
    can_clone_databases =  False
    test_db_allows_multiple_connections = False
    supports_unspecified_pk =  False
    can_return_columns_from_insert = False

    supports_column_check_constraints = False
    supports_table_check_constraints = False
    can_introspect_check_constraints = False

    interprets_empty_strings_as_nulls = True



from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.backends.base.client import BaseDatabaseClient
from django.db.backends.base.creation import BaseDatabaseCreation
from django.core.exceptions import ImproperlyConfigured
from django.utils.asyncio import async_unsafe

from .introspection import DatabaseIntrospection
from .features import DatabaseFeatures
from .schema import DatabaseSchemaEditor
from .operations import DatabaseOperations
from .cursor import CursorWrapper
from .creation import DatabaseCreation
from .validation import DatabaseValidation

import iris as Database


Database.Warning = type("StandardError", (object,), {})
Database.Error = type("StandardError", (object,), {})

Database.InterfaceError = type("Error", (object,), {})

Database.DatabaseError = type("Error", (object,), {})
Database.DataError = type("DatabaseError", (object,), {})
Database.OperationalError = type("DatabaseError", (object,), {})
Database.IntegrityError = type("DatabaseError", (object,), {})
Database.InternalError = type("DatabaseError", (object,), {})
Database.ProgrammingError = type("DatabaseError", (object,), {})
Database.NotSupportedError = type("DatabaseError", (object,), {})


def ignore(*args, **kwargs):
    pass


class DatabaseClient(BaseDatabaseClient):
    runshell = ignore


class DatabaseWrapper(BaseDatabaseWrapper):
    vendor = 'intersystems'
    display_name = 'InterSystems IRIS'

    data_types = {
        'AutoField': 'INTEGER AUTO_INCREMENT',
        'BigAutoField': 'BIGINT AUTO_INCREMENT',
        'BinaryField': 'LONG BINARY',
        'BooleanField': 'BIT',
        'CharField': 'VARCHAR(%(max_length)s)',
        'DateField': 'DATE',
        'DateTimeField': 'TIMESTAMP',
        'DecimalField': 'NUMERIC(%(max_digits)s, %(decimal_places)s)',
        'DurationField': 'BIGINT',
        'FileField': 'VARCHAR(%(max_length)s)',
        'FilePathField': 'VARCHAR(%(max_length)s)',
        'FloatField': 'DOUBLE PRECISION',
        'IntegerField': 'INTEGER',
        'BigIntegerField': 'BIGINT',
        'IPAddressField': 'CHAR(15)',
        'GenericIPAddressField': 'CHAR(39)',
        'JSONField': 'VARCHAR(32768)',
        'OneToOneField': 'INTEGER',
        'PositiveBigIntegerField': 'BIGINT',
        'PositiveIntegerField': 'INTEGER',
        'PositiveSmallIntegerField': 'SMALLINT',
        'SlugField': 'VARCHAR(%(max_length)s)',
        'SmallAutoField': 'SMALLINT AUTO_INCREMENT',
        'SmallIntegerField': 'SMALLINT',
        # 'TextField': 'LONG',
        # Stream not supported yet by db-api
        'TextField': 'VARCHAR(255)',
        'TimeField': 'TIME(6)',
        'UUIDField': 'CHAR(32)',
    }

    operators = {
        'exact': '= %s',
        'iexact': "LIKE %s ESCAPE '\\'",
        'contains': "LIKE %s ESCAPE '\\'",
        'icontains': "LIKE %s ESCAPE '\\'",
        # 'regex': 'REGEXP %s',
        # 'iregex': "REGEXP '(?i)' || %s",
        'gt': '> %s',
        'gte': '>= %s',
        'lt': '< %s',
        'lte': '<= %s',
        'startswith': "%STARTSWITH %s",
        'endswith': "LIKE %s ESCAPE '\\'",
        'istartswith': "%STARTSWITH %s",
        'iendswith': "LIKE %s ESCAPE '\\'",

    }
    Database = Database

    _commit = ignore
    _rollback = ignore
    _close = ignore
    _savepoint = ignore
    _savepoint_commit = ignore
    _savepoint_rollback = ignore
    _set_autocommit = ignore

    SchemaEditorClass = DatabaseSchemaEditor

    client_class = DatabaseClient
    creation_class = DatabaseCreation
    features_class = DatabaseFeatures
    introspection_class = DatabaseIntrospection
    ops_class = DatabaseOperations
    validation_class = DatabaseValidation

    def get_connection_params(self):
        settings_dict = self.settings_dict

        conn_params = {
            'username': None,
            'password': None,
        }
        if settings_dict['USER']:
            conn_params['username'] = settings_dict['USER']
        if settings_dict['PASSWORD']:
            conn_params['password'] = settings_dict['PASSWORD']

        if 'CONNECTION_STRING' in settings_dict:
            conn_params['connectionstr'] = settings_dict['CONNECTION_STRING']
        else:
            conn_params = {
                'hostname': 'localhost',
                'port': 1972,
                'namespace': 'USER',
                **conn_params,
            }
            if settings_dict['HOST']:
                conn_params['hostname'] = settings_dict['HOST']
            if settings_dict['PORT']:
                conn_params['port'] = settings_dict['PORT']
            if settings_dict['NAME']:
                conn_params['namespace'] = settings_dict['NAME']
            if 'NAMESPACE' in settings_dict:
                conn_params['namespace'] = settings_dict['NAMESPACE']

            if (
                not conn_params['hostname'] or
                not conn_params['port'] or
                not conn_params['namespace']
            ):
                raise ImproperlyConfigured(
                    "settings.DATABASES is improperly configured. "
                    "Please supply the HOST, PORT and NAME"
                )

        if (
            not conn_params['username'] or
            not conn_params['password']
        ):
            raise ImproperlyConfigured(
                "settings.DATABASES is improperly configured. "
                "Please supply the USER and PASSWORD"
            )

        return conn_params

    @async_unsafe
    def get_new_connection(self, conn_params):
        return Database.connect(**conn_params)

    def init_connection_state(self):
        cursor = self.connection.cursor()
        # cursor.callproc('%SYSTEM_SQL.Util_SetOption', ['SELECTMODE', 1])
        # cursor.callproc('%SYSTEM.SQL_SetSelectMode', [1])

    @async_unsafe
    def create_cursor(self, name=None):
        cursor = self.connection.cursor()
        return CursorWrapper(cursor)

    def is_usable(self):
        try:
            with self.connection.cursor() as cursor:
                cursor.execute('SELECT 1')
        except:
            return False
        else:
            return True



from decimal import Decimal

from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db import NotSupportedError

class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):
    sql_alter_column_type = "ALTER COLUMN %(column)s %(type)s"
    sql_alter_column_null = "ALTER COLUMN %(column)s NULL"

    def quote_value(self, value):
        if isinstance(value, bool):
            return str(int(value))
        elif isinstance(value, (Decimal, float, int)):
            return str(value)
        elif isinstance(value, str):
            return "'%s'" % value.replace("\'", "\'\'")
        elif value is None:
            return "NULL"
        return value

    def prepare_default(self, value):
        return self.quote_value(value)

    def table_sql(self, model):
        if '.' in model._meta.db_table:
            raise NotSupportedError(
                "Invalid table name '%s'" % (model._meta.db_table)
            )
        return super().table_sql(model)


from setuptools import setup
import os

thelibFolder = os.path.dirname(os.path.realpath(__file__))
requirementPath = thelibFolder + '/requirements.txt'

requirements, dependency_links = [], []
if os.path.isfile(requirementPath):
    with open('./requirements.txt') as f:
        for line in f.read().splitlines():
            if line.startswith('http'):
                dependency_links.append(line)
            else:
                requirements.append(line)

setup(install_requires=requirements,
        dependency_links=dependency_links)



from django.db.models.expressions import Exists

def exists_as_intersystems(self, compiler, connection, template=None, **extra_context):
    template = "(SELECT COUNT(*) FROM (%(subquery)s))"
    return self.as_sql(compiler, connection, template, **extra_context)

Exists.as_intersystems = exists_as_intersystems




import imp
from django.conf import settings
from django.db.backends.base.operations import BaseDatabaseOperations
from django.utils import timezone
from itertools import chain
from datetime import datetime

import traceback

class DatabaseOperations(BaseDatabaseOperations):
    def quote_name(self, name):
        if name.startswith('"') and name.endswith('"'):
            return name  # Quoting once is enough.
        return '"%s"' % name

    def last_insert_id(self, cursor, table_name, pk_name):
        cursor.execute("SELECT TOP 1 %(pk)s FROM %(table)s ORDER BY %(pk)s DESC" % {
            'table': table_name,
            'pk': pk_name,
        })
        return cursor.fetchone()[0]

    def sql_flush(self, style, tables, *, reset_sequences=False, allow_cascade=False):
        if tables and allow_cascade:
            # Simulate TRUNCATE CASCADE by recursively collecting the tables
            # referencing the tables to be flushed.
            tables = set(chain.from_iterable(
                self._references_graph(table) for table in tables))
        sql = ['%s %s %s;' % (
            style.SQL_KEYWORD('DELETE'),
            style.SQL_KEYWORD('FROM'),
            style.SQL_FIELD(self.quote_name(table))
        ) for table in tables]
        if reset_sequences:
            sequences = [{'table': table} for table in tables]
            sql.extend(self.sequence_reset_by_name_sql(style, sequences))
        return sql

    def no_limit_value(self):
        return None

    def limit_offset_sql(self, low_mark, high_mark):
        # print("limit_offset_sql")
        limit, offset = self._get_limit_offset_params(low_mark, high_mark)
        return ' '.join(sql for sql in (
            # ('LIMIT %d' % limit) if limit else None,
            # ('OFFSET %d' % offset) if offset else None,
        ) if sql)

    def adapt_datetimefield_value(self, value):
        # print('adapt_datetimefield_value', value)
        # print(*traceback.format_stack(), sep='\n')
        if value is None:
            return None
        # Expression values are adapted by the database.
        if hasattr(value, 'resolve_expression'):
            return value

        value = int(value.timestamp() * 1000000)
        if value >= 0:
            value += 2 ** 60
        else:
            value += -(2 ** 61 * 3)
        
        return str(value)
        # return str(value).split("+")[0]

    def get_db_converters(self, expression):
        converters = super().get_db_converters(expression)
        internal_type = expression.output_field.get_internal_type()
        if internal_type == "DateTimeField":
            converters.append(self.convert_datetimefield_value)
        return converters

    def convert_datetimefield_value(self, value, expression, connection):
        # print('convert_datetimefield_value', value, expression)
        if isinstance(value, int):
            if value > 0:
                value -= 2 ** 60
            else:
                value -= -(2 ** 61 * 3)
            value = value / 1000000
            value = datetime.fromtimestamp(value)
        return value
